"""Top-level package for Py3DEP."""

from __future__ import annotations

from importlib.metadata import PackageNotFoundError, version

from py3dep import exceptions
from py3dep.exceptions import (
    InputTypeError,
    InputValueError,
    MissingColumnError,
    MissingCRSError,
    NoOutletError,
)
from py3dep.print_versions import show_versions
from py3dep.py3dep import (
    add_elevation,
    check_3dep_availability,
    elevation_bycoords,
    elevation_bygrid,
    elevation_profile,
    get_dem,
    get_dem_vrt,
    get_map,
    query_3dep_sources,
    static_3dep_dem,
)
from py3dep.utils import deg2mpm, fill_depressions

try:
    __version__ = version("py3dep")
except PackageNotFoundError:
    __version__ = "999"

__all__ = [
    # Functions
    "fill_depressions",
    "get_map",
    "check_3dep_availability",
    "query_3dep_sources",
    "deg2mpm",
    "elevation_bycoords",
    "elevation_bygrid",
    "elevation_profile",
    "static_3dep_dem",
    "get_dem",
    "get_dem_vrt",
    "add_elevation",
    "show_versions",
    # Exceptions
    "MissingColumnError",
    "MissingCRSError",
    "NoOutletError",
    "InputTypeError",
    "InputValueError",
    "exceptions",
    "__version__",
]



"""Get data from 3DEP database."""

# pyright: reportGeneralTypeIssues=false
from __future__ import annotations

import contextlib
import itertools
from typing import TYPE_CHECKING, Any, Literal, Sequence, Union, cast, overload

import geopandas as gpd
import numpy as np
import pandas as pd
import pyproj
import rasterio
import rioxarray as rxr
import xarray as xr
from rasterio import RasterioIOError
from shapely import LineString, MultiLineString, MultiPolygon, Polygon, ops
from shapely import box as shapely_box

import async_retriever as ar
import pygeoutils as geoutils
from py3dep import utils
from py3dep.exceptions import (
    InputTypeError,
    InputValueError,
    MissingCRSError,
    ServiceUnavailableError,
)
from pygeoogc import WMS, ArcGISRESTful, ServiceURL, ZeroMatchedError
from pygeoogc import utils as ogc_utils

if TYPE_CHECKING:
    from pathlib import Path

    CRSTYPE = Union[int, str, pyproj.CRS]

MAX_PIXELS = 8000000
LAYERS = [
    "DEM",
    "Hillshade Gray",
    "Aspect Degrees",
    "Aspect Map",
    "GreyHillshade_elevationFill",
    "Hillshade Multidirectional",
    "Slope Map",
    "Slope Degrees",
    "Hillshade Elevation Tinted",
    "Height Ellipsoidal",
    "Contour 25",
    "Contour Smoothed 25",
]
__all__ = [
    "get_map",
    "elevation_bygrid",
    "elevation_bycoords",
    "elevation_profile",
    "check_3dep_availability",
    "query_3dep_sources",
    "static_3dep_dem",
    "get_dem",
    "get_dem_vrt",
    "add_elevation",
]


@overload
def get_map(
    layers: str,
    geometry: Polygon | MultiPolygon | tuple[float, float, float, float],
    resolution: int,
    geo_crs: CRSTYPE = ...,
    crs: CRSTYPE = ...,
) -> xr.DataArray: ...


@overload
def get_map(
    layers: list[str],
    geometry: Polygon | MultiPolygon | tuple[float, float, float, float],
    resolution: int,
    geo_crs: CRSTYPE = ...,
    crs: CRSTYPE = ...,
) -> xr.Dataset: ...


def get_map(
    layers: str | list[str],
    geometry: Polygon | MultiPolygon | tuple[float, float, float, float],
    resolution: int,
    geo_crs: CRSTYPE = 4326,
    crs: CRSTYPE = 4326,
) -> xr.Dataset | xr.DataArray:
    """Access dynamic layer of `3DEP <https://www.usgs.gov/core-science-systems/ngp/3dep>`__.

    The 3DEP service has multi-resolution sources, so depending on the user
    provided resolution the data is resampled on server-side based
    on all the available data sources. The following layers are available:

    - ``DEM``
    - ``Hillshade Gray``
    - ``Aspect Degrees``
    - ``Aspect Map``
    - ``GreyHillshade_elevationFill``
    - ``Hillshade Multidirectional``
    - ``Slope Map``
    - ``Slope Degrees``
    - ``Hillshade Elevation Tinted``
    - ``Height Ellipsoidal``
    - ``Contour 25``
    - ``Contour Smoothed 25``

    Parameters
    ----------
    layers : str or list of str
        A valid 3DEP layer or a list of them.
    geometry : Polygon, MultiPolygon, or tuple
        A shapely Polygon or a bounding box of the form ``(west, south, east, north)``.
    resolution : int
        The target resolution in meters. The width and height of the output are computed in
        pixels based on the geometry bounds and the given resolution.
    geo_crs : str, int, or pyproj.CRS, optional
        The spatial reference system of the input geometry, defaults to ``EPSG:4326``.
    crs : str, int, or pyproj.CRS, optional
        The spatial reference system to be used for requesting the data, defaults to
        ``EPSG:4326``. Valid values are ``EPSG:4326``, ``EPSG:3576``, ``EPSG:3571``,
        ``EPSG:3575``, ``EPSG:3857``, ``EPSG:3572``, ``CRS:84``, ``EPSG:3573``,
        and ``EPSG:3574``.

    Returns
    -------
    xarray.DataArray or xarray.Dataset
        The requested topographic data as an ``xarray.DataArray`` or ``xarray.Dataset``.
    """
    _layers = list(layers) if isinstance(layers, (list, tuple)) else [layers]
    invalid = [lyr for lyr in _layers if lyr not in LAYERS]
    if invalid:
        raise InputValueError("layers", LAYERS, ", ".join(invalid))

    if "DEM" in _layers:
        _layers[_layers.index("DEM")] = "None"

    _layers = [f"3DEPElevation:{lyr}" for lyr in _layers]

    wms_url = ServiceURL().wms.nm_3dep

    valid_crs = ogc_utils.valid_wms_crs(wms_url)

    if len(valid_crs) == 0:
        raise ServiceUnavailableError(wms_url)

    if ogc_utils.validate_crs(crs).lower() not in valid_crs:
        raise InputValueError("crs", valid_crs)

    _geometry = geoutils.geo2polygon(geometry, geo_crs, crs)
    wms = WMS(wms_url, layers=_layers, outformat="image/tiff", crs=crs, validation=False)
    r_dict = wms.getmap_bybox(_geometry.bounds, resolution, box_crs=crs, max_px=MAX_PIXELS)

    try:
        ds = geoutils.gtiff2xarray(r_dict, _geometry, crs)
    except RasterioIOError as ex:
        raise ServiceUnavailableError(wms_url) from ex
    valid_layers = wms.get_validlayers()
    return utils.rename_layers(ds, list(valid_layers))


def static_3dep_dem(
    geometry: Polygon | MultiPolygon | tuple[float, float, float, float],
    crs: CRSTYPE,
    resolution: int = 10,
) -> xr.DataArray:
    """Get DEM data at specific resolution from 3DEP.

    Notes
    -----
    In contrast to ``get_map`` function, this function only gets DEM data at
    specific resolution, namely 10 m, 30 m, and 60 m. However, this function
    is faster. This function is intended for cases where only need DEM at a
    specific resolution is required and for the other requests ``get_map``
    should be used.

    Parameters
    ----------
    geometry : Polygon, MultiPolygon, or tuple of length 4
        Geometry to get DEM within. It can be a polygon or a boundong box
        of form (xmin, ymin, xmax, ymax).
    crs : int, str, of pyproj.CRS
        CRS of the input geometry.
    resolution : int, optional
        Target DEM source resolution in meters, defaults to 10 m which is the highest
        resolution available over the US. Available options are 10, 30, and 60.

    Returns
    -------
    xarray.DataArray
        The request DEM at the specified resolution.
    """
    base_url = "https://prd-tnm.s3.amazonaws.com/StagedProducts/Elevation"
    url = {
        10: f"{base_url}/13/TIFF/USGS_Seamless_DEM_13.vrt",
        30: f"{base_url}/1/TIFF/USGS_Seamless_DEM_1.vrt",
        60: f"{base_url}/2/TIFF/USGS_Seamless_DEM_2.vrt",
    }
    if resolution not in url:
        raise InputValueError("resolution", list(url))

    dem = rxr.open_rasterio(url[resolution])
    dem = cast("xr.DataArray", dem)
    if "band" in dem.dims:
        dem = dem.squeeze("band")
    poly = geoutils.geo2polygon(geometry, crs, dem.rio.crs)
    dem = dem.rio.clip_box(*poly.bounds)
    if isinstance(geometry, (Polygon, MultiPolygon)):
        dem = dem.rio.clip([poly])
    dem = dem.where(dem > dem.rio.nodata, drop=False)
    dem = dem.rio.write_nodata(np.nan)
    dem.attrs.update({"units": "meters", "vertical_datum": "NAVD88", "vertical_resolution": 0.001})
    dem.name = "elevation"
    return dem


def get_dem(
    geometry: Polygon | MultiPolygon | tuple[float, float, float, float],
    resolution: int,
    crs: CRSTYPE = 4326,
) -> xr.DataArray:
    """Get DEM data at any resolution from 3DEP.

    Notes
    -----
    This function is a wrapper of ``static_3dep_dem`` and ``get_map`` functions.
    Since ``static_3dep_dem`` is much faster, if the requested resolution is 10 m,
    30 m, or 60 m, ``static_3dep_dem`` will be used. Otherwise, ``get_map``
    will be used.

    Parameters
    ----------
    geometry : Polygon, MultiPolygon, or tuple of length 4
        Geometry to get DEM within. It can be a polygon or a boundong box
        of form (xmin, ymin, xmax, ymax).
    resolution : int
        Target DEM source resolution in meters.
    crs : str, int, or pyproj.CRS, optional
        The spatial reference system of the input geometry, defaults to ``EPSG:4326``.

    Returns
    -------
    xarray.DataArray
        DEM at the specified resolution in meters and 4326 CRS.
    """
    if np.isclose(resolution, (10, 30, 60)).any():
        dem = static_3dep_dem(geometry, crs, resolution)
    else:
        dem = get_map("DEM", geometry, resolution, crs)
    dem = dem.astype("f4")
    dem.attrs.update({"units": "m", "vertical_datum": "NAVD88"})
    dem.name = "elevation"
    return dem


def add_elevation(
    ds: xr.DataArray | xr.Dataset,
    resolution: int | None = None,
    x_dim: str = "x",
    y_dim: str = "y",
    mask: xr.DataArray | None = None,
) -> xr.Dataset:
    """Add elevation data to a dataset as a new variable.

    Parameters
    ----------
    ds : xarray.DataArray or xarray.Dataset
        The dataset to add elevation data to. It must contain
        CRS information.
    resolution : float, optional
        Target DEM source resolution in meters, defaults ``None``, i.e.,
        the resolution of the input ``ds`` will be used.
    x_dim : str, optional
        Name of the x-coordinate dimension in ``ds``, defaults to ``x``.
    y_dim : str, optional
        Name of the y-coordinate dimension in ``ds``, defaults to ``y``.
    mask : xarray.DataArray, optional
        A mask to apply to the elevation data, defaults to ``None``.

    Returns
    -------
    xarray.Dataset
        The dataset with ``elevation`` variable added.
    """
    if not isinstance(ds, (xr.DataArray, xr.Dataset)):
        raise InputTypeError("ds", "xarray.DataArray or xarray.Dataset")

    ds_crs = ds.rio.crs
    if ds_crs is None:
        raise MissingCRSError

    if isinstance(ds, xr.DataArray):
        name = ds.name or "data"
        ds = ds.to_dataset(name=name, promote_attrs=True)
    else:
        ds = ds.copy()
    ds = ds.rio.set_spatial_dims(x_dim=x_dim, y_dim=y_dim)
    if ds_crs.is_projected:
        ds_proj = ds
        crs_proj = ds_crs
    else:
        ds_proj = ds.rio.reproject(5070)
        crs_proj = 5070
    ds_bounds = ds_proj.rio.bounds()
    if resolution is None:
        resolution = int(ds_proj.rio.resolution()[0])
    bounds = shapely_box(*ds_bounds).buffer(3 * resolution, join_style=2, cap_style=2)
    bounds = geoutils.geometry_reproject(bounds, crs_proj, 4326)

    dims_order = (y_dim, x_dim)
    elev = (
        get_dem(bounds.bounds, resolution).rename({"x": x_dim, "y": y_dim}).transpose(*dims_order)
    )
    if ds_crs != elev.rio.crs:
        elev = elev.rio.reproject(ds_crs)
    elev = geoutils.xarray_geomask(elev, ds.rio.bounds(), ds_crs)
    ds["elevation"] = (dims_order, elev.rio.reproject_match(ds, resampling=1).values)
    ds["elevation"] = ds["elevation"].rio.write_crs(
        ds.rio.crs, grid_mapping_name=ds.rio.grid_mapping
    )
    if mask is not None:
        ds["elevation"] = ds["elevation"].where(mask)
    ds = cast("xr.Dataset", ds)
    return ds


def get_dem_vrt(
    bbox: tuple[float, float, float, float],
    resolution: int,
    vrt_path: Path | str,
    tiff_dir: Path | str = "cache",
    crs: CRSTYPE = 4326,
) -> None:
    """Get DEM data at any resolution from 3DEP and save it as a VRT file.

    Parameters
    ----------
    bbox : tuple of length 4
        The boundong box of form (xmin, ymin, xmax, ymax).
    resolution : int
        Target DEM source resolution in meters.
    vrt_path : str or pathlib.Path
        Path to the output VRT file.
    tiff_dir : str or pathlib.Path, optional
        Path to the directory to save the downloaded TIFF file, defaults
        to ``./cache``.
    crs : str, int, or pyproj.CRS, optional
        The spatial reference system of ``bbox``, defaults to ``EPSG:4326``.
    """
    wms_url = ServiceURL().wms.nm_3dep
    bounds = geoutils.geometry_reproject(bbox, crs, 4326)
    fname = WMS(
        wms_url, layers="3DEPElevation:None", outformat="image/tiff", crs=4326, validation=False
    ).getmap_bybox(bounds, resolution, max_px=MAX_PIXELS, tiff_dir=tiff_dir)
    return geoutils.gtiff2vrt(fname, vrt_path)


def elevation_bygrid(
    xcoords: Sequence[float],
    ycoords: Sequence[float],
    crs: CRSTYPE,
    resolution: int,
    depression_filling: bool = False,
) -> xr.DataArray:
    """Get elevation from DEM data for a grid.

    This function is intended for getting elevations for a gridded dataset.

    Parameters
    ----------
    xcoords : list
        List of x-coordinates of a grid.
    ycoords : list
        List of y-coordinates of a grid.
    crs : str, int, or pyproj.CRS or pyproj.CRS
        The spatial reference system of the input grid,
        defaults to ``EPSG:4326``.
    resolution : int
        The accuracy of the output, defaults to 10 m which is the highest
        available resolution that covers CONUS. Note that higher resolution
        increases computation time so chose this value with caution.
    depression_filling : bool, optional
        Fill depressions before sampling using
        `pyflwdir <https://deltares.github.io/pyflwdir>`__ package,
        defaults to ``False``.

    Returns
    -------
    xarray.DataArray
        Elevations of the input coordinates as a ``xarray.DataArray``.
    """
    pts_crs = ogc_utils.validate_crs(crs)
    bbox = (
        gpd.GeoSeries(gpd.points_from_xy(*zip(*itertools.product(xcoords, ycoords)), crs=pts_crs))
        .to_crs(5070)
        .buffer(2 * resolution)
        .total_bounds
    )

    dem = get_dem(tuple(bbox), resolution, 5070)
    dem = dem.rio.reproject(pts_crs)

    if depression_filling:
        dem = utils.fill_depressions(dem)

    return dem.interp(x=list(xcoords), y=list(ycoords))


class ElevationByCoords:
    """Elevation model class.

    Parameters
    ----------
    crs : str, int, or pyproj.CRS, optional
        Coordinate reference system of the input coordinates, defaults to ``EPSG:4326``.
    coords : list of tuple
        List of coordinates.
    source : str, optional
        Elevation source, defaults to ``tep``. Valid sources are: ``tnm``, and ``tep``.
    """

    def __init__(
        self,
        coords: list[tuple[float, float]] | tuple[float, float],
        crs: CRSTYPE = 4326,
        source: str = "tep",
    ) -> None:
        self.coords = geoutils.coords_list(coords)
        self.crs = ogc_utils.validate_crs(crs)
        self.coords_gs = gpd.GeoSeries(gpd.points_from_xy(*zip(*self.coords)), crs=self.crs)
        valid_sources = ("tnm", "tep")
        self.source = source
        if self.source not in valid_sources:
            raise InputValueError("source", valid_sources)

    @property
    def values(self) -> list[float]:
        """Return list of elevations in meters."""
        if self.source == "tep":
            return self.tep()
        return self.tnm()

    def tnm(self) -> list[float]:
        """Return list of elevations in meters."""
        pts = self.coords_gs.to_crs(4326)
        kwds = [
            {"params": {"units": "Meters", "x": f"{x:.6f}", "y": f"{y:.6f}"}}
            for x, y in zip(pts.x, pts.y)
        ]
        resp = ar.retrieve_json([ServiceURL().restful.nm_pqs] * len(kwds), kwds, max_workers=5)
        resp = cast("list[dict[str, Any]]", resp)
        return [float(r["value"]) for r in resp]

    def tep(self) -> list[float]:
        """Get elevation from 10-m resolution 3DEP."""
        url = "/".join(
            (
                "https://prd-tnm.s3.amazonaws.com/StagedProducts/Elevation",
                "13/TIFF/USGS_Seamless_DEM_13.vrt",
            )
        )
        with rasterio.open(url) as src:
            points_proj = self.coords_gs.to_crs(src.crs)
            elev = np.array(list(src.sample(zip(points_proj.x, points_proj.y)))).ravel()
        return elev.tolist()


@overload
def elevation_bycoords(
    coords: tuple[float, float], crs: CRSTYPE = ..., source: Literal["tep", "tnm"] = ...
) -> float: ...


@overload
def elevation_bycoords(
    coords: list[tuple[float, float]],
    crs: CRSTYPE = ...,
    source: Literal["tep", "tnm"] = ...,
) -> list[float]: ...


def elevation_bycoords(
    coords: tuple[float, float] | list[tuple[float, float]],
    crs: CRSTYPE = 4326,
    source: Literal["tep", "tnm"] = "tep",
) -> float | list[float]:
    """Get elevation for a list of coordinates.

    Parameters
    ----------
    coords : tuple or list of tuple
        Coordinates of target location(s), e.g., ``[(x, y), ...]``.
    crs : str, int, or pyproj.CRS or pyproj.CRS, optional
        Spatial reference (CRS) of coords, defaults to ``EPSG:4326``.
    source : str, optional
        Data source to be used, default to ``tep``. Supported sources are
        ``tnm`` (using The National Map's Bulk Point
        Query Service with 10 m resolution) and ``tep`` (using 3DEP's static DEM VRTs
        at 10 m resolution). The ``tnm`` and ``tep`` sources are more accurate since they
        use the 1/3 arc-second DEM layer from 3DEP service but it is limited to the US.
        Note that ``tnm`` is bit unstable. It's recommended to use ``tep`` unless 10-m
        resolution accuracy is not necessary.

    Returns
    -------
    float or list of float
        Elevation in meter.
    """
    _crs = crs.to_string() if isinstance(crs, pyproj.CRS) else crs
    service = ElevationByCoords(crs=_crs, coords=coords, source=source)
    if len(service.coords) == 1:
        return service.values[0]
    return service.values


def elevation_profile(
    lines: LineString | MultiLineString,
    spacing: float,
    crs: CRSTYPE = 4326,
) -> xr.DataArray:
    """Get the elevation profile along a line at a given uniform spacing.

    .. note::

        This function converts the line to a spline and then calculates the elevation
        along the spline at a given uniform spacing using 10-m resolution DEM from 3DEP.

    Parameters
    ----------
    lines : LineString or MultiLineString
        Line segment(s) to be profiled. If its type is ``MultiLineString``,
        it will be converted to a single ``LineString`` and if this operation
        fails, an ``InputTypeError`` will be raised.
    spacing : float
        Spacing between the sample points along the line in meters.
    crs : str, int, or pyproj.CRS, optional
        Spatial reference System (CRS) of ``lines``, defaults to ``EPSG:4326``.

    Returns
    -------
    xarray.DataArray
        Elevation profile with dimension ``z`` and three coordinates: ``x``, ``y``,
        and ``distance``. The ``distance`` coordinate is the distance from the start
        of the line in meters.
    """
    if not isinstance(lines, (LineString, MultiLineString)):
        raise InputTypeError("lines", "LineString or MultiLineString")

    if isinstance(lines, MultiLineString):
        path = ops.linemerge(lines)
        if not isinstance(path, LineString):
            raise InputTypeError("lines", "MultiLineString that can be merged to LineString")
    else:
        path = lines

    crs_prj = 5070
    geom_proj = geoutils.geometry_reproject(path, crs, crs_prj)
    npts = int(np.ceil(geom_proj.length / spacing))
    geom_proj = geoutils.smooth_linestring(geom_proj, 0.1, npts)
    elev_list = elevation_bycoords(list(zip(*geom_proj.xy)), crs=crs_prj, source="tep")
    elevation = xr.DataArray(
        elev_list,
        dims="z",
        coords={"z": range(len(elev_list))},
        attrs={"source": "10-m DEM from 3DEP"},
    )
    geom = geoutils.geometry_reproject(geom_proj, crs_prj, crs)
    x, y = geom.xy
    elevation["x"], elevation["y"] = ("z", list(x)), ("z", list(y))
    distance = np.hypot(np.diff(x), np.diff(y)).cumsum()
    distance = np.insert(distance, 0, 0)
    elevation["distance"] = ("z", distance)
    elevation["distance"].attrs = {"units": "m", "long_name": "Distance from start"}
    return elevation


def check_3dep_availability(
    bbox: tuple[float, float, float, float], crs: CRSTYPE = 4326
) -> dict[str, bool | str]:
    """Query 3DEP's resolution availability within a bounding box.

    This function checks availability of 3DEP's at the following resolutions:
    1 m, 3 m, 5 m, 10 m, 30 m, 60 m, and topobathy (integrated topobathymetry).

    Parameters
    ----------
    bbox : tuple
        Bounding box as tuple of ``(min_x, min_y, max_x, max_y)``.
    crs : str, int, or pyproj.CRS or pyproj.CRS, optional
        Spatial reference (CRS) of ``bbox``, defaults to ``EPSG:4326``.

    Returns
    -------
    dict
        ``True`` if bbox intersects 3DEP elevation for each available resolution.
        Keys are the supported resolutions and values are their availability.
        If the query fails due to any reason, the value will be ``Failed``.
        If necessary, you can try again later until there is no ``Failed`` value.

    Examples
    --------
    >>> import py3dep
    >>> bbox = (-69.77, 45.07, -69.31, 45.45)
    >>> py3dep.check_3dep_availability(bbox)
    {'1m': True, '3m': False, '5m': False, '10m': True, '30m': True, '60m': False, 'topobathy': False}
    """
    if not isinstance(bbox, Sequence) or len(bbox) != 4:
        raise InputTypeError("bbox", "a tuple of length 4")

    res_layers = {
        "1m": 18,
        "3m": 19,
        "5m": 20,
        "10m": 21,
        "30m": 22,
        "60m": 23,
        "topobathy": 30,
    }
    base_url = ServiceURL().restful.nm_3dep_index
    urls = [f"{base_url}/{lyr}/query" for lyr in res_layers.values()]

    geom_query = ogc_utils.esri_query(bbox, crs, 4326)
    payload = {
        **geom_query,
        "spatialRel": "esriSpatialRelIntersects",
        "returnGeometry": "false",
        "returnIdsOnly": "false",
        "returnCountOnly": "true",
        "f": "json",
    }

    resps = ar.retrieve_json(
        urls,
        [{"params": payload}] * len(urls),
        max_workers=len(urls),
        raise_status=False,
    )
    resps = cast("list[dict[str, Any]]", resps)

    def _check(r: dict[str, Any] | None) -> bool | str:
        if r is None or "error" in r:
            return "Failed"
        return bool(r.get("count"))

    avail = {res: _check(r) for res, r in zip(res_layers, resps)}
    failed = [res for res, r in avail.items() if r == "Failed"]
    if failed:
        [
            ar.delete_url_cache(f"{base_url}/{res_layers[res]}/query", params=payload)
            for res in failed
        ]
    return avail


def query_3dep_sources(
    bbox: tuple[float, float, float, float],
    crs: CRSTYPE = 4326,
    res: str | list[str] | None = None,
) -> gpd.GeoDataFrame:
    """Query 3DEP's data sources within a bounding box.

    This function queries the availability of the underlying data that 3DEP uses
    at the following resolutions:
    1 m, 3 m, 5 m, 10 m, 30 m, 60 m, and topobathy (integrated topobathymetry).

    Parameters
    ----------
    bbox : tuple
        Bounding box as tuple of ``(min_x, min_y, max_x, max_y)``.
    crs : str, int, or pyproj.CRS or pyproj.CRS, optional
        Spatial reference (CRS) of bbox, defaults to ``EPSG:4326``.
    res : str, list of str, optional
        Resolution to query, defaults to ``None``, i.e., all resolutions.
        Available resolutions are: ``1m``, ``3m``, ``5m``, ``10m``, ``30m``,
        ``60m``, and ``topobathy``.

    Returns
    -------
    geopandas.GeoDataFrame
        Polygon(s) representing the 3DEP data sources at each resolution.
        Resolutions are given in the ``dem_res`` column.

    Examples
    --------
    >>> import py3dep
    >>> bbox = (-69.77, 45.07, -69.31, 45.45)
    >>> src = py3dep.query_3dep_sources(bbox)
    >>> src.groupby("dem_res")["OBJECTID"].count().to_dict()
    {'10m': 16, '1m': 4, '30m': 8}
    >>> src = py3dep.query_3dep_sources(bbox, res="1m")
    >>> src.groupby("dem_res")["OBJECTID"].count().to_dict()
    {'1m': 4}
    """
    if not isinstance(bbox, Sequence) or len(bbox) != 4:
        raise InputTypeError("bbox", "a tuple of length 4")

    res_layers = {
        "1m": 18,
        "3m": 19,
        "5m": 20,
        "10m": 21,
        "30m": 22,
        "60m": 23,
        "topobathy": 30,
    }
    if res is None:
        layers = res_layers
    elif isinstance(res, str) and res in res_layers:
        layers = {res: res_layers[res]}
    elif isinstance(res, (list, tuple)) and all(r in res_layers for r in res):
        layers = {r: res_layers[r] for r in res}
    else:
        raise InputValueError("res", list(res_layers))

    def _check(lyr: int) -> gpd.GeoDataFrame | None:
        client = ArcGISRESTful(ServiceURL().restful.nm_3dep_index, lyr, outformat="json")
        with contextlib.suppress(ZeroMatchedError):
            oids = client.oids_bygeom(bbox, crs)
            return geoutils.json2geodf(client.get_features(oids))
        return None

    return (  # pyright: ignore[reportReturnType]
        gpd.GeoDataFrame(  # pyright: ignore[reportCallIssue]
            pd.concat({res: _check(lyr) for res, lyr in layers.items()}).reset_index(  # pyright: ignore[reportCallIssue,reportArgumentType]
                level=1, drop=True
            ),
            crs=4326,
        )
        .reset_index()
        .rename(columns={"index": "dem_res"})
    )



"""Utilities for Py3DEP."""

from __future__ import annotations

import functools
import heapq
import warnings
from typing import TYPE_CHECKING, Any, Callable, Literal, Sequence, TypeVar, Union, overload

import numpy as np
import xarray as xr
from numpy.typing import NDArray

from py3dep.exceptions import InputTypeError, InputValueError, NoOutletError

if TYPE_CHECKING:
    import pyproj

    CRSTYPE = Union[int, str, pyproj.CRS]


__all__ = ["deg2mpm", "fill_depressions"]


try:
    from numba import config as numba_config
    from numba import njit, prange

    numba_config.THREADING_LAYER = "workqueue"
    has_numba = True
except ImportError:
    has_numba = False
    prange = range

    T = TypeVar("T")
    Func = Callable[..., T]

    def njit(
        signature_or_function: str | Func[T], nogil: bool = True, parallel: bool = False
    ) -> Callable[[Func[T]], Func[T]]:
        def decorator_njit(func: Func[T]) -> Func[T]:
            @functools.wraps(func)
            def wrapper_decorator(*args: tuple[Any, ...], **kwargs: dict[str, Any]) -> T:
                return func(*args, **kwargs)

            return wrapper_decorator

        return decorator_njit


FloatArray = NDArray[np.float32]
BoolArray = NDArray[np.bool_]
IntArray = NDArray[np.uint32]
DataArray = TypeVar("DataArray", FloatArray, xr.DataArray)


@njit(
    "boolean[:, ::1](f4[:, ::1], boolean[:, ::1], boolean[:, ::1], optional(uint32[::1]), optional(f4))",
    nogil=True,
    parallel=True,
)
def _get_queued(
    elevtn: FloatArray,
    not_done: BoolArray,
    structure: BoolArray,
    idxs_pit: IntArray | None,
    elv_max: np.float32 | None,
) -> BoolArray:
    """Get queued cells for depression filling."""
    nrow, ncol = elevtn.shape
    if idxs_pit is not None:
        # initiate queue with outlet cells
        queued = np.full((nrow, ncol), False, dtype=np.bool_)
        for idx in idxs_pit:
            queued.flat[idx] = True
        return queued
    # initiate queue with edge cells
    s = np.where(structure.ravel())[0]
    queued = not_done.copy()
    for r in prange(0, nrow):
        for c in prange(0, ncol):
            if ~not_done[r, c] or r in (0, nrow - 1) or c in (0, ncol - 1):
                continue
            a0 = not_done[slice(r - 1, r + 2), slice(c - 1, c + 2)].ravel()
            if np.all(a0[s]):
                queued[r, c] = False
    if elv_max is not None:
        queued = np.logical_and(queued, elevtn <= elv_max)
        if not np.any(queued):
            raise NoOutletError
    return queued


@njit(
    "f4[:, ::1](f4[:, ::1], unicode_type, optional(uint32[::1]), f4, f4, optional(f4), i4)",
    nogil=True,
)
def _fill_depressions(
    elevtn: FloatArray,
    outlets: Literal["edge", "min"],
    idxs_pit: IntArray | None,
    nodata: np.float32,
    max_depth: np.float32,
    elv_max: np.float32 | None,
    connectivity: np.uint8,
) -> FloatArray:
    """Fill local depressions in elevation data based on Wang and Liu (2006)."""
    nrow, ncol = elevtn.shape
    delv = elevtn.copy()
    done = np.isnan(elevtn) if np.isnan(nodata) else np.isclose(elevtn, nodata)

    struct = np.full((3, 3), True, dtype=np.bool_)
    if connectivity == 4:
        struct[0, 0], struct[-1, -1] = False, False
        struct[0, -1], struct[-1, 0] = False, False

    queued = _get_queued(elevtn, ~done, struct, idxs_pit, elv_max)

    # queue contains (elevation, row, col)
    # boundary is included to favor non-boundary cells over boundary
    # cells with same elevation
    q = [(np.float32(0), np.uint32(0), np.uint32(0)) for _ in range(0)]
    for r, c in zip(*np.where(queued)):
        heapq.heappush(q, (np.float32(elevtn[r, c]), np.uint32(r), np.uint32(c)))
    heapq.heapify(q)

    # restrict queue to global edge minimum (single outlet)
    if outlets == "min":
        q = [heapq.heappop(q)]
        queued = np.full((nrow, ncol), False, dtype=np.bool_)
        queued[q[0][-2], q[0][-1]] = True

    # loop over cells and neighbors with ascending cell elevation.
    drs, dcs = np.where(struct)
    drs, dcs = drs - 1, dcs - 1
    while q:
        z0, r0, c0 = heapq.heappop(q)
        for dr, dc in zip(drs, dcs):
            r = r0 + dr
            c = c0 + dc
            if not (0 <= r < nrow and 0 <= c < ncol) or done[r, c]:
                continue
            z1 = elevtn[r, c]
            # local depression if dz > 0
            dz = z0 - z1
            # if positive max_depth: don't fill when dz > max_depth
            if max_depth >= 0 and dz >= max_depth:
                if not queued[r, c]:
                    heapq.heappush(q, (np.float32(z1), np.uint32(r), np.uint32(c)))
                    queued[r, c] = True
                continue
            if dz > 0:
                delv[r, c] += dz
            # add to queue if not already in queue
            if ~queued[r, c]:
                heapq.heappush(q, (np.float32(delv[r, c]), np.uint32(r), np.uint32(c)))
                queued[r, c] = True
            done[r, c] = True
    return delv


def fill_depressions(
    elevtn: DataArray,
    outlets: Literal["edge", "min"] = "min",
    idxs_pit: IntArray | None = None,
    nodata: float = np.nan,
    max_depth: float = -1.0,
    elv_max: float | None = None,
    connectivity: Literal[4, 8] = 8,
) -> DataArray:
    """Fill local depressions in elevation data based on Wang and Liu (2006).

    .. note::

        This function is based on the ``fill_depressions`` function from the
        `pyflwdir <https://github.com/Deltares/pyflwdir>`__ package. This function
        improves the performance of the original function by a factor of up to 2 and
        adds more input checks. Additionally, it works with ``xarray.DataArray`` objects.

        Outlets are assumed to occur at the edge of valid elevation cells ``outlets='edge'``;
        at the lowest valid edge cell to create one single outlet ``outlets='min'``;
        or at user provided outlet cells ``idxs_pit``.

        Depressions elsewhere are filled based on its lowest pour point elevation.
        If the pour point depth is larger than the maximum pour point depth ``max_depth``
        a pit is set at the depression local minimum elevation.

        Wang, L., & Liu, H. (2006). https://doi.org/10.1080/13658810500433453

    Parameters
    ----------
    elevtn: numpy.ndarray or xarray.DataArray
        elevation raster as a 2D ``numpy.ndarray`` or ``xarray.DataArray``.
    outlets: {"edge", "min}, optional
        Initial basin outlet(s) at the edge of all cells ('edge')
        or only the minimum elevation edge cell ('min'; default)
    idxs_pit: 1D array of int, optional
        Linear indices of outlet cells, in any, defaults to None.
    nodata: float, optional
        nodata value, defaults to ``numpy.nan``.
    max_depth: float, optional
        Maximum pour point depth. Depressions with a larger pour point
        depth are set as pit. A negative value (default) equals an infitely
        large pour point depth causing all depressions to be filled.
        Defaults to -1.0.
    elv_max, float, optional
        Maximum elevation for outlets, only in combination with ``outlets='edge'``.
        By default ``None``.
    connectivity: {4, 8}, optional
        Number of neighboring cells to consider, defaults to 8.

    Returns
    -------
    elevtn_out: numpy.ndarray
        Depression filled elevation with type float32.
    """
    if not has_numba:
        warnings.warn(
            "Numba not installed. Using very slow pure python version.", UserWarning, stacklevel=2
        )
    if not isinstance(elevtn, (np.ndarray, xr.DataArray)):
        raise InputTypeError("elevtn", "2D numpy.ndarray or xarray.DataArray")
    if elevtn.ndim != 2:
        raise InputTypeError("elevtn", "2D numpy.ndarray or xarray.DataArray")
    if outlets not in ("edge", "min"):
        raise InputValueError("outlets", ("edge", "min"))
    if connectivity not in (4, 8):
        raise InputValueError("connectivity", (4, 8))
    if idxs_pit is not None:  # noqa: SIM102
        if not isinstance(elevtn, np.ndarray) and idxs_pit.ndim != 1:
            raise InputTypeError("idxs_pit", "1D numpy.ndarray")
    _elevtn = np.asarray(elevtn, dtype=np.float32)
    _idxs_pit = None if idxs_pit is None else np.asarray(idxs_pit, dtype=np.uint32)
    _nodata = np.float32(nodata)
    _max_depth = np.float32(max_depth)
    _elv_max = None if elv_max is None else np.float32(elv_max)
    _connectivity = np.uint8(connectivity)
    corrected = _fill_depressions(
        _elevtn, outlets, _idxs_pit, _nodata, _max_depth, _elv_max, _connectivity
    )
    if isinstance(elevtn, xr.DataArray):
        return elevtn.copy(data=corrected)
    return corrected


def deg2mpm(slope: xr.DataArray) -> xr.DataArray:
    """Convert slope from degrees to meter/meter.

    Parameters
    ----------
    slope : xarray.DataArray
        Slope in degrees.

    Returns
    -------
    xarray.DataArray
        Slope in meter/meter. The name is set to ``slope`` and the ``units`` attribute
        is set to ``m/m``.
    """
    with xr.set_options(keep_attrs=True):
        if hasattr(slope, "_FillValue"):
            nodata = slope.attrs["_FillValue"]
        elif hasattr(slope, "nodatavals"):
            _nodata = slope.attrs["nodatavals"]
            nodata = _nodata[0] if isinstance(_nodata, Sequence) else _nodata
        else:
            nodata = np.nan
        slope = slope.where(slope != nodata, drop=False)

        def to_mpm(da: xr.DataArray) -> xr.DataArray:
            """Convert slope from degrees to meter/meter."""
            return xr.apply_ufunc(
                lambda x: np.tan(np.deg2rad(x)),
                da,
                vectorize=True,
            )

        slope = to_mpm(slope).compute()
        slope.attrs["nodatavals"] = (np.nan,)
        if hasattr(slope, "_FillValue"):
            slope.attrs["_FillValue"] = np.nan
        slope.name = "slope"
        slope.attrs["units"] = "m/m"
    return slope


@overload
def rename_layers(ds: xr.DataArray, valid_layers: list[str]) -> xr.DataArray: ...


@overload
def rename_layers(ds: xr.Dataset, valid_layers: list[str]) -> xr.Dataset: ...


def rename_layers(
    ds: xr.DataArray | xr.Dataset, valid_layers: list[str]
) -> xr.DataArray | xr.Dataset:
    """Rename layers in a dataset."""
    rename = {lyr: lyr.split(":")[-1].replace(" ", "_").lower() for lyr in valid_layers}
    rename.update({"3DEPElevation:None": "elevation"})

    if isinstance(ds, xr.DataArray):
        ds.name = rename[str(ds.name)]
    else:
        ds = ds.rename({n: rename[str(n)] for n in ds})
    return ds



from __future__ import annotations

import pytest
from shapely import Polygon

import py3dep
from py3dep import InputTypeError, InputValueError

DEF_CRS = 4326
GEOM = Polygon(
    [[-69.77, 45.07], [-69.31, 45.07], [-69.31, 45.45], [-69.77, 45.45], [-69.77, 45.07]]
)


def test_wrong_bbox():
    bbox = (1, 2, 3)
    with pytest.raises(InputTypeError, match="tuple of length 4"):
        _ = py3dep.check_3dep_availability(bbox)


def test_wrong_layer():
    geom = GEOM.bounds
    with pytest.raises(InputValueError, match="DEM"):
        _ = py3dep.get_map("None", geom, 1000)


def test_wrong_crs():
    geom = GEOM.bounds
    with pytest.raises(InputValueError, match="crs"):
        _ = py3dep.get_map("DEM", geom, 1000, DEF_CRS, "ESRI:102003")



"""Customized Py3DEP exceptions."""

from __future__ import annotations

import async_retriever as ar
import pygeoogc as ogc


class InputTypeError(ar.InputTypeError):
    """Exception raised when a function argument type is invalid.

    Parameters
    ----------
    arg : str
        Name of the function argument
    valid_type : str
        The valid type of the argument
    example : str, optional
        An example of a valid form of the argument, defaults to None.
    """


class InputValueError(ar.InputValueError):
    """Exception raised for invalid input.

    Parameters
    ----------
    inp : str
        Name of the input parameter
    valid_inputs : tuple
        List of valid inputs
    given : str, optional
        The given input, defaults to None.
    """


class MissingColumnError(Exception):
    """Exception raised when a required column is missing from a dataframe.

    Parameters
    ----------
    missing : list
        List of missing columns.
    """

    def __init__(self, missing: list[str]) -> None:
        self.message = f"The following columns are missing:\n{', '.join(missing)}"
        super().__init__(self.message)

    def __str__(self) -> str:
        """Return the error message."""
        return self.message


class NoOutletError(Exception):
    """Exception raised when no outlet is found in a DEM."""

    def __init__(self) -> None:
        self.message = "No outlet was foundm, please provide an outlet point or adjust ``elv_max``."
        super().__init__(self.message)

    def __str__(self) -> str:
        """Return the error message."""
        return self.message


class MissingCRSError(Exception):
    """Exception raised when input is missing CRS."""

    def __init__(self) -> None:
        self.message = "The input is missing CRS."
        super().__init__(self.message)

    def __str__(self) -> str:
        """Return the error message."""
        return self.message


class ServiceUnavailableError(ogc.ServiceUnavailableError):
    """Exception raised when the service is not available.

    Parameters
    ----------
    url : str
        The server url
    """


class InputRangeError(Exception):
    """Exception raised when input is out of range.

    Parameters
    ----------
    inp : str
        Name of the input parameter
    valid_range : str
        The valid range of the input
    """

    def __init__(self, inp: str, valid_range: str) -> None:
        self.message = f"{inp} is out of range. Valid range is {valid_range}"
        super().__init__(self.message)

    def __str__(self) -> str:
        """Return the error message."""
        return self.message



"""Configuration for pytest."""

from __future__ import annotations

import pytest
from click.testing import CliRunner


@pytest.fixture
def runner():
    """Return a CliRunner."""
    return CliRunner()


@pytest.fixture(autouse=True)
def _add_standard_imports(doctest_namespace):
    """Add py3dep namespace for doctest."""
    import py3dep

    doctest_namespace["py3dep"] = py3dep



"""Command-line interface for Py3DEP."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Literal, TypeVar, cast

import click
import geopandas as gpd
import pandas as pd

from py3dep import py3dep
from py3dep.exceptions import InputTypeError, MissingColumnError, MissingCRSError
from py3dep.py3dep import LAYERS

if TYPE_CHECKING:
    DFType = TypeVar("DFType", pd.DataFrame, gpd.GeoDataFrame)


def get_target_df(tdf: DFType, req_cols: list[str]) -> DFType:
    """Check if all required columns exists in the dataframe.

    It also re-orders the columns based on ``req_cols`` order.
    """
    missing = [c for c in req_cols if c not in tdf]
    if missing:
        raise MissingColumnError(missing)
    return tdf[req_cols]  # pyright: ignore[reportReturnType]


CONTEXT_SETTINGS = {"help_option_names": ["-h", "--help"]}

save_arg = click.option(
    "-s",
    "--save_dir",
    default="topo_3dep",
    type=click.Path(exists=False),
    help=" ".join(
        (
            "Path to a directory to save the requested files.",
            "Extension for the outputs is either `.nc` for geometry or `.csv` for coords.",
        )
    ),
)


@click.group(context_settings=CONTEXT_SETTINGS)
def cli() -> None:
    """Command-line interface for Py3DEP."""


@cli.command("coords", context_settings=CONTEXT_SETTINGS)
@click.argument("fpath", type=click.Path(exists=True))
@click.option(
    "-q",
    "--query_source",
    default="tep",
    type=click.Choice(["tnm", "tep"], case_sensitive=False),
    help="Source of the elevation data: The National Map (tnm) or 3DEP (tep).",
)
@save_arg
def coords(
    fpath: Path,
    query_source: Literal["tep", "tnm"] = "tep",
    save_dir: str | Path = "topo_3dep",
) -> None:
    """Retrieve topographic data for a list of coordinates.

    \b
    FPATH: Path to a csv file with two columns named ``lon`` and ``lat``.

    \b
    Examples:
        $ cat coords.csv
        lon,lat
        -122.2493328,37.8122894
        $ py3dep coords coords.csv -q tep -s topo_dir
    """  # noqa: D301
    fpath = Path(fpath)
    elev = get_target_df(pd.read_csv(fpath), ["lon", "lat"])

    count = "1 point" if len(elev) == 1 else f"{len(elev)} points"
    click.echo(f"Found coordinates of {count} in {fpath.resolve()}. Retrieving ... ")

    coords_list = list(elev.itertuples(index=False, name=None))
    coords_list = cast("list[tuple[float, float]]", coords_list)
    elev["elevation"] = py3dep.elevation_bycoords(coords_list, 4326, query_source)

    Path(save_dir).mkdir(parents=True, exist_ok=True)
    elev.astype("f4").to_csv(Path(save_dir, f"{fpath.stem}_elevation.csv"))
    click.echo("Done.")


@cli.command("geometry", context_settings=CONTEXT_SETTINGS)
@click.argument("fpath", type=click.Path(exists=True))
@click.option(
    "-l",
    "--layers",
    multiple=True,
    default=["DEM"],
    type=click.Choice(LAYERS, case_sensitive=True),
    help="Target topographic data layers",
)
@save_arg
def geometry(
    fpath: Path,
    layers: str | list[str] = "DEM",
    save_dir: str | Path = "topo_3dep",
) -> None:
    """Retrieve topographic data within geometries.

    \b
    FPATH: Path to a shapefile (.shp) or geopackage (.gpkg) file.
    This file must have three columns and contain a ``crs`` attribute:
        - ``id``: Feature identifiers that py3dep uses as the output netcdf/csv filenames.
        - ``res``: Target resolution in meters.
        - ``geometry``: A Polygon or MultiPloygon.

    \b
    Examples:
        $ py3dep geometry ny_geom.gpkg -l "Slope Map" -l DEM -s topo_dir
    """  # noqa: D301
    fpath = Path(fpath)
    if fpath.suffix not in (".shp", ".gpkg"):
        raise InputTypeError("file", ".shp or .gpkg")

    target_df = gpd.read_file(fpath)
    if target_df.crs is None:
        raise MissingCRSError
    crs = target_df.crs.to_string()

    target_df = get_target_df(target_df, ["id", "res", "geometry"])
    args_list = ((g, r, Path(save_dir, f"{i}.nc")) for i, r, g in target_df.itertuples(index=False))

    count = "1 geometry" if len(target_df) == 1 else f"{len(target_df)} geometries"
    click.echo(f"Found {count} in {fpath.resolve()}.")

    Path(save_dir).mkdir(parents=True, exist_ok=True)
    with click.progressbar(
        args_list, label="Getting topographic data from 3DEP", length=len(target_df)
    ) as bar:
        for geo, res, f in bar:
            py3dep.get_map(layers, geo, res, geo_crs=crs, crs=4326).to_netcdf(f)
    click.echo("Done.")



from __future__ import annotations

import io
import shutil
import subprocess
from pathlib import Path

import geopandas as gpd
import numpy as np
import pandas as pd
import pyproj
import pytest
import rioxarray as rxr
import xarray as xr
from shapely import MultiLineString, Polygon, ops

import py3dep
from py3dep.cli import cli
from pygeoogc import utils

DEF_CRS = 4326
ALT_CRS = 3857
GEOM = Polygon(
    [[-69.77, 45.07], [-69.31, 45.07], [-69.31, 45.45], [-69.77, 45.45], [-69.77, 45.07]]
)
LINE = MultiLineString(
    [
        [[-69.77, 45.07], [-69.31, 45.07]],
        [[-69.31, 45.07], [-69.31, 45.45]],
        [[-69.31, 45.45], [-69.77, 45.45]],
    ]
)
LYR = "Slope Degrees"
SMALL = 1e-3

has_gdal = subprocess.getstatusoutput("gdalinfo --version")[0] == 0


def assert_close(a: float, b: float, rtol: float = 1e-3) -> None:
    assert np.isclose(a, b, rtol=rtol).all()


def test_profile():
    ep = py3dep.elevation_profile(LINE, 1000)
    epm = py3dep.elevation_profile(ops.linemerge(LINE), 1000)
    expected = 264.6431
    assert_close(ep.mean().item(), expected)
    assert_close(epm.mean().item(), expected)


def test_getmap():
    layers = ["DEM", LYR]
    ds = py3dep.get_map(layers, GEOM.bounds, 1000, geo_crs=DEF_CRS, crs=ALT_CRS)
    dem_30 = py3dep.get_map(layers[0], GEOM, 30, geo_crs=DEF_CRS, crs=ALT_CRS)
    dem_1e3 = py3dep.get_map(layers[0], GEOM, 1000, geo_crs=DEF_CRS, crs=ALT_CRS)
    fpath = Path("dem_30.tif")
    dem_30.rio.to_raster(fpath)
    dem_30 = rxr.open_rasterio(fpath)
    assert sorted(ds.keys()) == ["elevation", "slope_degrees"]
    assert_close(dem_30.mean().item(), dem_1e3.mean().item(), 0.5)
    dem_30.close()
    fpath.unlink()


def test_dem():
    expected = 295.686
    ds = py3dep.get_dem(GEOM, 10)
    assert_close(ds.mean().item(), expected)
    ds = py3dep.get_dem(GEOM, 15)
    assert_close(ds.mean().item(), expected)


@pytest.mark.skipif(not has_gdal, reason="GDAL is not installed")
def test_dem_vrt():
    expected = 295.6827
    py3dep.get_dem_vrt(GEOM.bounds, 30, "cache/dem.vrt")
    ds = rxr.open_rasterio("cache/dem.vrt").squeeze(drop=True)
    assert_close(ds.mean().item(), expected)


@pytest.mark.speedup
def test_fill_depressions():
    ds = py3dep.get_dem(GEOM.bounds, 1000)
    ds = py3dep.fill_depressions(ds)
    assert_close(ds.mean().item(), 296.9658)


@pytest.mark.parametrize(
    ("source", "expected"),
    [("tnm", 356.093), ("tep", 356.139)],
)
def test_bycoords(source, expected):
    coords = [(-7766049.664788851, 5691929.739021257)]
    dem = py3dep.elevation_bycoords(coords * 101, pyproj.CRS(ALT_CRS), source=source)
    assert_close(sum(set(dem)), expected)


def test_deg2mpm():
    slope = py3dep.get_map(LYR, GEOM, 1000)
    slope = py3dep.deg2mpm(slope)
    assert_close(slope.mean().item(), 0.0505)


def test_grid():
    crs = (
        "+proj=lcc +lat_1=25 +lat_2=60 +lat_0=42.5 +lon_0=-100 +x_0=0 +y_0=0 +ellps=WGS84 +units=m"
    )
    geom = utils.match_crs(GEOM, DEF_CRS, crs)
    xmin, ymin, xmax, ymax = geom.bounds
    res = 1000
    gx = np.arange(xmin, xmax, res)
    gy = np.arange(ymin, ymax, res)
    elev = py3dep.elevation_bygrid(tuple(gx), tuple(gy), crs, res)
    elev_fill = py3dep.elevation_bygrid(tuple(gx), tuple(gy), crs, res, depression_filling=True)
    assert_close((elev_fill - elev).sum().item(), 9096.3853)


def test_add_elev():
    crs = (
        "+proj=lcc +lat_1=25 +lat_2=60 +lat_0=42.5 +lon_0=-100 +x_0=0 +y_0=0 +ellps=WGS84 +units=m"
    )
    geom = utils.match_crs(GEOM, DEF_CRS, crs)
    xmin, ymin, xmax, ymax = geom.bounds
    res = 10e3
    gx = np.arange(xmin, xmax, res)
    gy = np.arange(ymin, ymax, res)
    da = xr.DataArray(dims=["y", "x"], coords={"y": gy, "x": gx})
    da = da.rio.write_crs(crs)
    ds = py3dep.add_elevation(da)
    assert_close(ds["elevation"].mean().item(), 291.3313)


def test_check_3dep_availability():
    avail = py3dep.check_3dep_availability(GEOM.bounds)
    assert avail["1m"]
    assert avail["10m"]
    assert avail["30m"]


def test_query_3dep_source():
    src = py3dep.query_3dep_sources(GEOM.bounds)
    res_all = src.groupby("dem_res")["OBJECTID"].count().to_dict()
    src = py3dep.query_3dep_sources(GEOM.bounds, res="1m")
    res_1m = src.groupby("dem_res")["OBJECTID"].count().to_dict()
    assert res_all == {"10m": 8, "1m": 4, "30m": 8}
    assert res_1m == {"1m": 4}


class TestCLI:
    """Test the command-line interface."""

    def test_geometry(self, runner):
        gdf = gpd.GeoDataFrame(
            {"id": "geo_test", "res": 1e3}, geometry=[GEOM], index=[0], crs=DEF_CRS
        )
        geo_gpkg = Path("nat_geo.gpkg")
        gdf.to_file(geo_gpkg)
        ret = runner.invoke(
            cli, ["geometry", str(geo_gpkg), "-l", "DEM", "-l", LYR, "-s", "geo_map"]
        )
        if geo_gpkg.is_dir():
            shutil.rmtree(geo_gpkg)
        else:
            geo_gpkg.unlink()
        assert ret.exit_code == 0
        assert "Found 1 geometry" in ret.output
        shutil.rmtree("geo_map")

    def test_coords(self, runner):
        df = pd.DataFrame(
            [[-69.77, 45.07], [-69.31, 45.07], [-69.31, 45.45]], columns=["lon", "lat"]
        )
        coord_csv = "coords.csv"
        df.to_csv(coord_csv)
        ret = runner.invoke(cli, ["coords", coord_csv, "-s", "geo_coords", "-q", "tep"])
        Path(coord_csv).unlink()
        assert ret.exit_code == 0
        assert "Found coordinates of 3 points" in ret.output
        shutil.rmtree("geo_coords")


def test_show_versions():
    f = io.StringIO()
    py3dep.show_versions(file=f)
    assert "SYS INFO" in f.getvalue()
